/*
 * BasicShapeGeneratorPlugin.java 4 september 2022
 *
 * Sweet Home 3D, Copyright (c) 2022 Emmanuel PUYBARET / eTeks <info@eteks.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package com.eteks.sweethome3d.plugin.basicshapegenerator;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.lang.ref.WeakReference;
import java.net.URL;
import java.util.Arrays;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.concurrent.Callable;

import javax.jnlp.BasicService;
import javax.jnlp.ServiceManager;
import javax.jnlp.UnavailableServiceException;
import javax.media.j3d.Appearance;
import javax.media.j3d.BranchGroup;
import javax.media.j3d.Material;
import javax.media.j3d.Shape3D;
import javax.media.j3d.Transform3D;
import javax.media.j3d.TransformGroup;
import javax.media.j3d.TransparencyAttributes;
import javax.swing.DefaultListCellRenderer;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JEditorPane;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSlider;
import javax.swing.JSpinner;
import javax.swing.KeyStroke;
import javax.swing.border.EmptyBorder;
import javax.swing.event.HyperlinkEvent;
import javax.swing.event.HyperlinkListener;
import javax.swing.undo.AbstractUndoableEdit;
import javax.swing.undo.UndoableEditSupport;
import javax.vecmath.Color3f;
import javax.vecmath.Point3f;
import javax.vecmath.Vector3f;

import com.eteks.sweethome3d.j3d.ModelManager;
import com.eteks.sweethome3d.j3d.OBJWriter;
import com.eteks.sweethome3d.model.CatalogPieceOfFurniture;
import com.eteks.sweethome3d.model.Content;
import com.eteks.sweethome3d.model.HomePieceOfFurniture;
import com.eteks.sweethome3d.model.InterruptedRecorderException;
import com.eteks.sweethome3d.model.RecorderException;
import com.eteks.sweethome3d.model.UserPreferences;
import com.eteks.sweethome3d.plugin.Plugin;
import com.eteks.sweethome3d.plugin.PluginAction;
import com.eteks.sweethome3d.swing.IconManager;
import com.eteks.sweethome3d.swing.NullableSpinner;
import com.eteks.sweethome3d.swing.NullableSpinner.NullableSpinnerLengthModel;
import com.eteks.sweethome3d.swing.SwingTools;
import com.eteks.sweethome3d.swing.SwingViewFactory;
import com.eteks.sweethome3d.tools.OperatingSystem;
import com.eteks.sweethome3d.tools.ResourceURLContent;
import com.eteks.sweethome3d.tools.TemporaryURLContent;
import com.eteks.sweethome3d.viewcontroller.HomeView;
import com.eteks.sweethome3d.viewcontroller.ThreadedTaskController;
import com.sun.j3d.utils.geometry.Box;
import com.sun.j3d.utils.geometry.Cone;
import com.sun.j3d.utils.geometry.Cylinder;
import com.sun.j3d.utils.geometry.GeometryInfo;
import com.sun.j3d.utils.geometry.NormalGenerator;
import com.sun.j3d.utils.geometry.Primitive;
import com.sun.j3d.utils.geometry.Sphere;

/**
 * A plug-in that creates a 3D shape from various basic shapes chosen by user.
 * @author Emmanuel Puybaret
 */
public class BasicShapeGeneratorPlugin extends Plugin {
  private enum ShapeType {TETRAHEDRON, PYRAMID, CUBE, CROSS, SPHERE, CYLINDER, CONE, HEXAGONAL_PRISM, DODECAHEDRON, TORUS};

  private JPanel    shapePanel;
  private JComboBox shapeComboBox;
  private JSlider   transparencySlider;
  private JSpinner  widthSpinner;

  @Override
  public PluginAction [] getActions() {
    return new PluginAction [] {new PluginAction(
            "com.eteks.sweethome3d.plugin.basicshapegenerator.ApplicationPlugin",
            "GENERATE_BASIC_SHAPE", getPluginClassLoader(), true) {
        @Override
        public void execute() {
          generateShape();
        }
      }
    };
  }

  /**
   * Creates a 3D shape from coordinates specified by user.
   */
  private void generateShape() {
    final ResourceBundle resource = ResourceBundle.getBundle(
        "com.eteks.sweethome3d.plugin.basicshapegenerator.ApplicationPlugin",
        Locale.getDefault(), getPluginClassLoader());
    final HomeView homeView = getHomeController().getView();

    try {
      // Ignore plug-in in protected Java Web Start environment
      ServiceManager.lookup("javax.jnlp.FileSaveService");
      // Use an uneditable editor pane to let user select text in dialog
      JEditorPane messagePane = new JEditorPane("text/html",
          resource.getString("BasicShapeGeneratorJavaWebStartInfo.message"));
      messagePane.setOpaque(false);
      messagePane.setEditable(false);
      try {
        // Lookup the javax.jnlp.BasicService object
        final BasicService service = (BasicService)ServiceManager.lookup("javax.jnlp.BasicService");
        // If basic service supports  web browser
        if (service.isWebBrowserSupported()) {
          // Add a listener that displays hyperlinks content in browser
          messagePane.addHyperlinkListener(new HyperlinkListener() {
            public void hyperlinkUpdate(HyperlinkEvent ev) {
              if (ev.getEventType() == HyperlinkEvent.EventType.ACTIVATED) {
                service.showDocument(ev.getURL());
              }
            }
          });
        }
      } catch (UnavailableServiceException ex) {
        // Too bad : service is unavailable
      }

      String title = resource.getString("basicShapeGeneratorJavaWebStartInfo.title");
      JOptionPane.showMessageDialog((JComponent)homeView, messagePane, title, JOptionPane.WARNING_MESSAGE);
      return;
    } catch (UnavailableServiceException ex) {
    }

    if (this.shapePanel == null) {
      // Reuse the same panel from a call to plug-in to the next to allow modifying values if they were wrong
      createComponents(resource);
      // Reset panel if unit changes
      getUserPreferences().addPropertyChangeListener(UserPreferences.Property.UNIT, new UserPreferencesChangeListener(this));
    }
    if (JOptionPane.showConfirmDialog((JComponent)homeView, this.shapePanel,
            resource.getString("shapeGeneratorDialog.title"), JOptionPane.OK_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE) == JOptionPane.OK_OPTION) {
      final ShapeType shapeType = (ShapeType)this.shapeComboBox.getSelectedItem();
      final float width = ((NullableSpinner.NullableSpinnerLengthModel)this.widthSpinner.getModel()).getLength();
      final float alpha = this.transparencySlider.getValue() / 255f;
      // Create a new 3D model in a threaded task
      Callable<Void> exportToObjTask = new Callable<Void>() {
          public Void call() throws RecorderException, IOException {
            final String shapeName = resource.getString(getShapeName(shapeType) + "Label");
            final Content modelContent = exportShape(shapeName, shapeType, width, alpha);
            homeView.invokeLater(new Runnable() {
                public void run() {
                  addNewPieceOfFurniture(modelContent, getShapeName(shapeType) + ".png",  shapeName, resource);
                }
              });
            return null;
          }
        };
      ThreadedTaskController.ExceptionHandler exceptionHandler =
          new ThreadedTaskController.ExceptionHandler() {
            public void handleException(Exception ex) {
              if (!(ex instanceof InterruptedRecorderException)) {
                showError((JComponent)homeView, resource, ex.getMessage());
                ex.printStackTrace();
              }
            }
          };
      new ThreadedTaskController(exportToObjTask,
          resource.getString("basicShapeGeneratorMessage"), exceptionHandler,
          getUserPreferences(), new SwingViewFactory()).executeTask(homeView);
    }
  }

  /**
   * Creates and layout the components shown in the dialog box of this plug-in.
   */
  private void createComponents(final ResourceBundle resource) {
    UserPreferences preferences = getUserPreferences();

    this.shapeComboBox = new JComboBox(ShapeType.values());
    this.shapeComboBox.setRenderer(new DefaultListCellRenderer() {
        @Override
        public Component getListCellRendererComponent(JList<?> list, Object value, int index, boolean isSelected, boolean cellHasFocus) {
          String shapeName = getShapeName((ShapeType)value);
          JLabel component = (JLabel)super.getListCellRendererComponent(list, resource.getString(shapeName + "Label"),
              index, isSelected, cellHasFocus);
          component.setIcon(IconManager.getInstance().getIcon(new ResourceURLContent(BasicShapeGeneratorPlugin.class, shapeName + ".png"), 16, list));
          component.setBorder(new EmptyBorder(0, 4, 0, 0));
          return component;
        }
      });
    this.shapeComboBox.setMaximumRowCount(Integer.MAX_VALUE);
    NullableSpinnerLengthModel widthModel = new NullableSpinner.NullableSpinnerLengthModel(preferences, 100f, preferences.getLengthUnit().getMinimumLength(), 10000f);
    this.widthSpinner = new NullableSpinner(widthModel);
    this.transparencySlider = new JSlider(0, 255);
    this.transparencySlider.setPaintTicks(true);
    this.transparencySlider.setMajorTickSpacing(17);
    this.transparencySlider.setValue(0);

    // Layout coordinates spinners and transparency slider in a panel with labels
    this.shapePanel = new JPanel(new GridBagLayout());
    int labelAlignment = OperatingSystem.isMacOSX()
        ? GridBagConstraints.LINE_END
        : GridBagConstraints.LINE_START;
    int standardGap = Math.round(5 * SwingTools.getResolutionScale());
    JLabel shapeLabel = new JLabel(resource.getString("shapeLabel.text"));
    this.shapePanel.add(shapeLabel,
        new GridBagConstraints(0, 0, 1, 1, 0, 0, labelAlignment,
            GridBagConstraints.NONE, new Insets(0, 0, standardGap, standardGap), 0, 0));
    this.shapePanel.add(this.shapeComboBox,
        new GridBagConstraints(1, 0, 1, 1, 0, 0, GridBagConstraints.LINE_START,
            GridBagConstraints.NONE, new Insets(0, 0, standardGap, 0), 0, 0));
    JLabel widthLabel = new JLabel(String.format(resource.getString("widthLabel.text"), preferences.getLengthUnit().getName()));
    this.shapePanel.add(widthLabel, new GridBagConstraints(
        0, 1, 1, 1, 0, 0, labelAlignment,
        GridBagConstraints.NONE, new Insets(0, 0, standardGap, standardGap), 0, 0));
    this.shapePanel.add(this.widthSpinner, new GridBagConstraints(
        1, 1, 1, 1, 0, 0, GridBagConstraints.LINE_START,
        GridBagConstraints.NONE, new Insets(0, 0, standardGap, 0), 0, 0));
    JLabel transparencyLabel = new JLabel(resource.getString("transparencyLabel.text"));
    this.shapePanel.add(transparencyLabel, new GridBagConstraints(
        0, 2, 1, 1, 0, 0, labelAlignment,
        GridBagConstraints.NONE, new Insets(0, 0, 0, standardGap), 0, 0));
    this.shapePanel.add(this.transparencySlider, new GridBagConstraints(
        1, 2, 1, 1, 0, 0, GridBagConstraints.LINE_START,
        GridBagConstraints.HORIZONTAL, new Insets(0, 0, 0, 0), 0, 0));
    JPanel transparencyLabelsPanel = new JPanel(new BorderLayout(20, 0));
    transparencyLabelsPanel.setOpaque(false);
    transparencyLabelsPanel.add(new JLabel(resource.getString("opaqueLabel.text")), BorderLayout.WEST);
    transparencyLabelsPanel.add(new JLabel(resource.getString("invisibleLabel.text")), BorderLayout.EAST);
    this.shapePanel.add(transparencyLabelsPanel, new GridBagConstraints(
        1, 3, 1, 1, 0, 0, GridBagConstraints.NORTH,
        GridBagConstraints.HORIZONTAL, new Insets(OperatingSystem.isWindows() ? 0 : -3, 0, 0, 0), 0, 0));

    if (!OperatingSystem.isMacOSX()) {
      shapeLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(resource.getString("shapeLabel.mnemonic")).getKeyCode());
      shapeLabel.setLabelFor(this.shapeComboBox);
      widthLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(resource.getString("widthLabel.mnemonic")).getKeyCode());
      widthLabel.setLabelFor(this.widthSpinner);
      transparencyLabel.setDisplayedMnemonic(KeyStroke.getKeyStroke(resource.getString("transparencyLabel.mnemonic")).getKeyCode());
      transparencyLabel.setLabelFor(this.transparencySlider);
    }
  }

  /**
   * Returns the localized name of the given shape type.
   */
  private String getShapeName(ShapeType shapeType) {
    return shapeType == ShapeType.HEXAGONAL_PRISM ? "hexagonalPrism" : shapeType.name().toLowerCase();
  }

  /**
   * Exports the shape from the given type.
   */
  private Content exportShape(String shapeName, ShapeType shapeType,
                              float width, float alpha) throws RecorderException {
    try {
      BranchGroup root = new BranchGroup();
      createShape(root, shapeType, width, alpha);

      if (root.numChildren() != 0) {
        File tempZipFile = OperatingSystem.createTemporaryFile(shapeName, ".zip");
        String objFile = shapeName + ".obj";
        OBJWriter.writeNodeInZIPFile(root, tempZipFile, 0, objFile, "Created by BasicShapeGenerator plug-in");
        return new TemporaryURLContent(new URL("jar:" + tempZipFile.toURI().toURL() + "!/" + objFile));
      } else {
        return null;
      }
    } catch (InterruptedIOException ex) {
      throw new InterruptedRecorderException("Export to OBJ interrupted");
    } catch (IOException ex) {
      throw new RecorderException("Couldn't export to OBJ", ex);
    }
  }

  /**
   * Adds to parent a 3D shape matching the coordinates if the shape isn't empty.
   */
  private void createShape(BranchGroup parent, ShapeType shapeType, float width, float alpha) {
    Appearance appearance = new Appearance();
    Material material = new Material(new Color3f(0.2f, 0.2f, 0.2f), new Color3f(), new Color3f(1.0f, 1.0f, 1.0f), new Color3f(), 0);
    try {
      appearance.setName("Material");
    } catch (NoSuchMethodError ex) {
      // Ignore appearance name with Java 3D < 1.4 where setName was added
    }
    appearance.setMaterial(material);
    if (alpha > 0) {
      appearance.setTransparencyAttributes(new TransparencyAttributes(TransparencyAttributes.NICEST, alpha));
    }

    // Values and rotation set to get width = 100
    switch (shapeType) {
      case TETRAHEDRON:
        parent.addChild(new Cone(width * 0.5774f, width * 0.8165f, Primitive.GENERATE_TEXTURE_COORDS, 3, 1, appearance));
        break;
      case PYRAMID:
        Transform3D rotation = new Transform3D();
        rotation.rotY(Math.PI / 4);
        TransformGroup rotationGroup = new TransformGroup(rotation);
        rotationGroup.addChild(new Cone(width * 0.7071f, width * 0.7071f, Primitive.GENERATE_TEXTURE_COORDS, 4, 1, appearance));
        parent.addChild(rotationGroup);
        break;
      case CUBE:
        parent.addChild(new Box(width / 2, width / 2, width / 2, Primitive.GENERATE_TEXTURE_COORDS, appearance));
        break;
      case CROSS:
        parent.addChild(new Box(width / 2, width / 6, width / 6, Primitive.GENERATE_TEXTURE_COORDS, appearance));
        parent.addChild(new Box(width / 6, width / 6, width / 2, Primitive.GENERATE_TEXTURE_COORDS, appearance));
        break;
      case SPHERE:
        parent.addChild(new Sphere(width / 2, Primitive.GENERATE_TEXTURE_COORDS | Primitive.GENERATE_NORMALS, 32, appearance));
        break;
      case CYLINDER:
        parent.addChild(new Cylinder(width / 2, width, Primitive.GENERATE_TEXTURE_COORDS | Primitive.GENERATE_NORMALS, 32, 1, appearance));
        break;
      case CONE:
        parent.addChild(new Cone(width / 2, width, Primitive.GENERATE_TEXTURE_COORDS | Primitive.GENERATE_NORMALS, 32, 1, appearance));
        break;
      case HEXAGONAL_PRISM:
        rotation = new Transform3D();
        rotation.rotY(Math.PI / 2);
        rotationGroup = new TransformGroup(rotation);
        rotationGroup.addChild(new Cylinder(width / 2, width, Primitive.GENERATE_TEXTURE_COORDS, 6, 1, appearance));
        parent.addChild(rotationGroup);
        break;
      case DODECAHEDRON:
        // From https://en.wikipedia.org/wiki/Regular_dodecahedron#Cartesian_coordinates
        float phi = (float)(1 + Math.sqrt(5)) / 2;
        Point3f [] vertices = {
            new Point3f(1, 1, 1),
            new Point3f(0, 1 / phi, phi),
            new Point3f(phi, 0, 1 / phi),
            new Point3f(1 / phi, phi, 0),
            new Point3f(-1, 1, 1),
            new Point3f(0, -1 / phi, phi),
            new Point3f(1, -1, 1),
            new Point3f(phi, 0, -1 / phi),
            new Point3f(1, 1, -1),
            new Point3f(-1 / phi, phi, 0),
            new Point3f(-phi, 0, 1 / phi),
            new Point3f(-1, -1, 1),
            new Point3f(1 / phi, -phi, 0),
            new Point3f(1, -1, -1),
            new Point3f(0, 1 / phi, -phi),
            new Point3f(-1, 1, -1),
            new Point3f(-1 / phi, -phi, 0),
            new Point3f(-phi, 0, -1 / phi),
            new Point3f(0, -1 / phi, -phi),
            new Point3f(-1, -1, -1)};
        int [] faceIndices = {
            0, 1, 5, 6, 2,
            0, 2, 7, 8, 3,
            0, 3, 9, 4, 1,
            1, 4, 10, 11, 5,
            2, 6, 12, 13, 7,
            3, 8, 14, 15, 9,
            5, 11, 16, 12, 6,
            7, 13, 18, 14, 8,
            9, 15, 17, 10, 4,
            19, 16, 11, 10, 17,
            19, 17, 15, 14, 18,
            19, 18, 13, 12, 16};
        GeometryInfo geometryInfo = new GeometryInfo(GeometryInfo.POLYGON_ARRAY);
        geometryInfo.setCoordinates(vertices);
        geometryInfo.setCoordinateIndices(faceIndices);
        int [] stripCounts = {5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5, 5};
        geometryInfo.setStripCounts(stripCounts);
        new NormalGenerator(0).generateNormals(geometryInfo);
        rotation = new Transform3D();
        rotation.rotX(0.55383);
        Transform3D scale = new Transform3D();
        scale.setScale(width * 0.30904f);
        scale.mul(rotation);
        TransformGroup scaleGroup = new TransformGroup(scale);
        scaleGroup.addChild(new Shape3D(geometryInfo.getIndexedGeometryArray(), appearance));
        parent.addChild(scaleGroup);
        break;
      case TORUS:
        float minorRadius = width / 8;
        float majorRadius = width / 2 - minorRadius;
        int minorDivisions = 16;
        int majorDivisions = 32;
        vertices = new Point3f [minorDivisions * majorDivisions * 4];
        Transform3D majorRotation = new Transform3D();
        Transform3D majorNextRotation = new Transform3D();
        for (int i = 0, n = 0; i < majorDivisions; i++) {
          majorRotation.rotY(Math.PI * 2 / majorDivisions * i);
          majorNextRotation.rotY(Math.PI * 2 / majorDivisions * ((i + 1) % majorDivisions));
          for (int j = 0; j < minorDivisions; j++) {
            float minorAngle = (float)(Math.PI * 2 / minorDivisions * j);
            float cosMinorAngle = (float)Math.cos(minorAngle);
            float sinMinorAngle = (float)Math.sin(minorAngle);
            float minorNextAngle = (float)(Math.PI * 2 / minorDivisions * ((j + 1) % minorDivisions));
            float cosMinorNextAngle = (float)Math.cos(minorNextAngle);
            float sinMinorNextAngle = (float)Math.sin(minorNextAngle);
            vertices [n] = new Point3f(majorRadius + minorRadius * cosMinorAngle, minorRadius * sinMinorAngle, 0);
            majorRotation.transform(vertices [n++]);
            vertices [n] = new Point3f(majorRadius + minorRadius * cosMinorAngle, minorRadius * sinMinorAngle, 0);
            majorNextRotation.transform(vertices [n++]);
            vertices [n] = new Point3f(majorRadius + minorRadius * cosMinorNextAngle, minorRadius * sinMinorNextAngle, 0);
            majorNextRotation.transform(vertices [n++]);
            vertices [n] = new Point3f(majorRadius + minorRadius * cosMinorNextAngle, minorRadius * sinMinorNextAngle, 0);
            majorRotation.transform(vertices [n++]);
          }
        }
        geometryInfo = new GeometryInfo(GeometryInfo.QUAD_ARRAY);
        geometryInfo.setCoordinates(vertices);
        new NormalGenerator(Math.PI / 2).generateNormals(geometryInfo);
        parent.addChild(new Shape3D(geometryInfo.getIndexedGeometryArray(), appearance));
        break;
    }
  }

  /**
   * Adds to home a new piece of furniture based on the given model.
   */
  private void addNewPieceOfFurniture(final Content modelContent,
                                      final String shapeIcon,
                                      final String shapeName,
                                      final ResourceBundle resource) {
    ModelManager.getInstance().loadModel(modelContent, new ModelManager.ModelObserver() {
        public void modelUpdated(BranchGroup modelRoot) {
          Vector3f size = ModelManager.getInstance().getSize(modelRoot);
          Content iconContent = new ResourceURLContent(BasicShapeGeneratorPlugin.class, shapeIcon);
          HomePieceOfFurniture piece = new HomePieceOfFurniture(new CatalogPieceOfFurniture(
                null, shapeName, null, iconContent, modelContent,
                size.x, size.z, size.y,
                0, true, null, System.getProperty("user.name"), true, null, null));
          piece.setX(size.x / 2);
          piece.setY(size.z / 2);
          piece.setElevation(0);

          UndoableEditSupport undoSupport = getUndoableEditSupport();
          undoSupport.beginUpdate();
          getHomeController().getFurnitureController().addFurniture(Arrays.asList(new HomePieceOfFurniture [] {piece}));
          undoSupport.postEdit(new AbstractUndoableEdit() {
              @Override
              public String getPresentationName() {
                return resource.getString("undoGenerateBasicShape");
              }
            });

          // End compound edit
          undoSupport.endUpdate();
        }

        public void modelError(Exception ex) {
          // Shouldn't happen since we import a model we just exported
        }
      });
  }

  /**
   * Shows a message error.
   */
  private void showError(Component parent,
                         ResourceBundle resource,
                         String messageDetail) {
    String messageFormat = resource.getString("basicShapeGeneratorError.message");
    JOptionPane.showMessageDialog(parent, String.format(messageFormat, messageDetail),
        resource.getString("basicShapeGeneratorError.title"), JOptionPane.ERROR_MESSAGE);
  }

  /**
   * Preferences property listener bound to this table with a weak reference to avoid
   * strong link between user preferences and this plug-in.
   */
  private static class UserPreferencesChangeListener implements PropertyChangeListener {
    private WeakReference<BasicShapeGeneratorPlugin>  plugin;

    public UserPreferencesChangeListener(BasicShapeGeneratorPlugin plugin) {
      this.plugin = new WeakReference<BasicShapeGeneratorPlugin>(plugin);
    }

    public void propertyChange(PropertyChangeEvent ev) {
      // If plugin was garbage collected, remove this listener from preferences
      BasicShapeGeneratorPlugin plugin = this.plugin.get();
      ((UserPreferences)ev.getSource()).removePropertyChangeListener(
          UserPreferences.Property.valueOf(ev.getPropertyName()), this);
      if (plugin != null
          && plugin.shapePanel != null) {
        plugin.shapePanel = null;
      }
    }
  }
}
